@@ -1,11 +1,23 @@
 /*
-    This program can be distributed under the terms of the GNU GPL.
-    See the file COPYING.
+    Copyright 2006,2007,2008 Martin Pärtel <martin.partel@gmail.com>
 
-    Author: Martin Prtel <martin.partel@gmail.com>
+    This file is part of bindfs.
+
+    bindfs is free software: you can redistribute it and/or modify
+    it under the terms of the GNU General Public License as published by
+    the Free Software Foundation, either version 2 of the License, or
+    (at your option) any later version.
+
+    bindfs is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+    GNU General Public License for more details.
+
+    You should have received a copy of the GNU General Public License
+    along with bindfs.  If not, see <http://www.gnu.org/licenses/>.
 
 
-    This file is based on the fusexmp_fh.c in FUSE 2.5.3,
+    This file is based on fusexmp_fh.c from FUSE 2.5.3,
     which had the following notice:
     ---
     FUSE: Filesystem in Userspace
@@ -78,6 +90,8 @@ static struct settings {
         CHMOD_IGNORE,
         CHMOD_DENY
     } chmod_policy;
+    
+    int chmod_allow_x;
 
     enum XAttrPolicy {
         XATTR_UNIMPLEMENTED,
@@ -91,6 +105,7 @@ static struct settings {
     gid_t *mirrored_members;
     int num_mirrored_members;
 
+    int ctime_from_mtime;
 } settings;
 
 
@@ -189,6 +204,12 @@ static int getattr_common(const char *procpath, struct stat *stbuf)
 {
     struct fuse_context *fc = fuse_get_context();
 
+    /* Copy mtime (file content modification time)
+       to ctime (inode/status change time)
+       if the user asked for that */
+    if (settings.ctime_from_mtime)
+        stbuf->st_ctime = stbuf->st_mtime;
+
     /* Report user-defined owner/group if specified */
     if (settings.new_uid != -1)
         stbuf->st_uid = settings.new_uid;
@@ -480,20 +501,45 @@ static int bindfs_link(const char *from, const char *to)
 
 static int bindfs_chmod(const char *path, mode_t mode)
 {
-    int res;
+    int file_execute_only = 0;
+    struct stat st;
+    mode_t diff = 0;
+
+    path = process_path(path);
+
+    if (settings.chmod_allow_x) {
+        /* Get the old permission bits and see which bits would change. */
+        if (lstat(path, &st) == -1)
+            return -errno;
+
+        if (S_ISREG(st.st_mode)) {
+            diff = (st.st_mode & 07777) ^ (mode & 07777);
+            file_execute_only = 1;
+        }
+    }
 
     switch (settings.chmod_policy) {
     case CHMOD_NORMAL:
-        path = process_path(path);
-
-        res = chmod(path, mode);
-        if (res == -1)
+        if (chmod(path, mode) == -1)
             return -errno;
-
         return 0;
     case CHMOD_IGNORE:
+        if (file_execute_only) {
+            diff &= 00111; /* See which execute bits were flipped.
+                              Forget about other differences. */
+            if (chmod(path, st.st_mode ^ diff) == -1)
+                return -errno;
+        }
         return 0;
     case CHMOD_DENY:
+        if (file_execute_only) {
+            if ((diff & 07666) == 0) {
+                /* Only execute bits have changed, so we can allow this. */
+                if (chmod(path, mode) == -1)
+                    return -errno;
+                return 0;
+            }
+        }
         return -EPERM;
     default:
         assert(0);
@@ -678,6 +724,8 @@ static int bindfs_fsync(const char *path, int isdatasync,
 static int bindfs_setxattr(const char *path, const char *name, const char *value,
                            size_t size, int flags)
 {
+    DPRINTF("setxattr %s %s=%s\n", path, name, value);
+    
     if (settings.xattr_policy == XATTR_READ_ONLY)
         return -EACCES;
 
@@ -697,6 +745,8 @@ static int bindfs_getxattr(const char *path, const char *name, char *value,
 {
     int res;
 
+    DPRINTF("getxattr %s %s\n", path, name);
+    
     path = process_path(path);
     /* fuse checks permissions for us */
 #ifdef HAVE_LGETXATTR
@@ -713,6 +763,8 @@ static int bindfs_listxattr(const char *path, char *list, size_t size)
 {
     int res;
 
+    DPRINTF("listxattr %s\n", path);
+    
     path = process_path(path);
     /* fuse checks permissions for us */
 #ifdef HAVE_LLISTXATTR
@@ -727,6 +779,8 @@ static int bindfs_listxattr(const char *path, char *list, size_t size)
 
 static int bindfs_removexattr(const char *path, const char *name)
 {
+    DPRINTF("removexattr %s %s\n", path, name);
+
     if (settings.xattr_policy == XATTR_READ_ONLY)
         return -EACCES;
 
@@ -800,6 +854,10 @@ static void print_usage(const char *progname)
            "                            all other users.\n"
            "  -n      --no-allow-other  Do not add -o allow_other to fuse options.\n"
            "\n"
+           "Permission bits:\n"
+           "  -p      --perms           Specify permissions, similar to chmod\n"
+           "                            e.g. og-x,og+rD,u=rwX,g+rw  or  0644,a+X\n"
+           "\n"
            "File creation policy:\n"
            "  --create-as-user          New files owned by creator (default for root). *\n"
            "  --create-as-mounter       New files owned by fs mounter (default for users).\n"
@@ -811,15 +869,16 @@ static void print_usage(const char *progname)
            "  --chmod-normal            Try to chmod the original files (the default).\n"
            "  --chmod-ignore            Have all chmods to fail silently.\n"
            "  --chmod-deny              Have all chmods fail with 'permission denied'.\n"
+           "  --chmod-allow-x           Allow changing file execute bits in any case.\n"
            "\n"
            "Extended attribute policy:\n"
            "  --xattr-none              Do not implement xattr operations.\n"
-           "  --xattr-ro                Read-only xattr operations (the default).\n"
-           "  --xattr-rw                Read-write xattr operations.\n"
+           "  --xattr-ro                Read-only xattr operations.\n"
+           "  --xattr-rw                Read-write xattr operations (the default).\n"
            "\n"
-           "Permission bits:\n"
-           "  -p      --perms           Specify permissions, similar to chmod\n"
-           "                            e.g. og-x,og+rD,u=rwX,g+rw  or  0644,a+X\n"
+           "Time-related:\n"
+           "  --ctime-from-mtime        Read file properties' change time\n"
+           "                            from file content modification time.\n"
            "\n"
            "FUSE options:\n"
            "  -o opt[,opt,...]          Mount options.\n"
@@ -855,9 +914,11 @@ enum OptionKey {
     OPTKEY_CHMOD_NORMAL,
     OPTKEY_CHMOD_IGNORE,
     OPTKEY_CHMOD_DENY,
+    OPTKEY_CHMOD_ALLOW_X,
     OPTKEY_XATTR_NONE,
     OPTKEY_XATTR_READ_ONLY,
-    OPTKEY_XATTR_READ_WRITE
+    OPTKEY_XATTR_READ_WRITE,
+    OPTKEY_CTIME_FROM_MTIME
 };
 
 static int process_option(void *data, const char *arg, int key,
@@ -895,6 +956,10 @@ static int process_option(void *data, const char *arg, int key,
         settings.chmod_policy = CHMOD_DENY;
         return 0;
 
+    case OPTKEY_CHMOD_ALLOW_X:
+        settings.chmod_allow_x = 1;
+        return 0;
+
     case OPTKEY_XATTR_NONE:
         settings.xattr_policy = XATTR_UNIMPLEMENTED;
         return 0;
@@ -905,6 +970,10 @@ static int process_option(void *data, const char *arg, int key,
         settings.xattr_policy = XATTR_READ_WRITE;
         return 0;
 
+    case OPTKEY_CTIME_FROM_MTIME:
+        settings.ctime_from_mtime = 1;
+	return 0;
+
     case OPTKEY_NONOPTION:
         if (!settings.mntsrc) {
             settings.mntsrc = arg;
@@ -968,9 +1037,11 @@ int main(int argc, char *argv[])
         OPT2("--chmod-normal", "chmod-normal", OPTKEY_CHMOD_NORMAL),
         OPT2("--chmod-ignore", "chmod-ignore", OPTKEY_CHMOD_IGNORE),
         OPT2("--chmod-deny", "chmod-deny", OPTKEY_CHMOD_DENY),
+        OPT2("--chmod-allow-x", "chmod-allow-x", OPTKEY_CHMOD_ALLOW_X),
         OPT2("--xattr-none", "xattr-none", OPTKEY_XATTR_NONE),
         OPT2("--xattr-ro", "xattr-ro", OPTKEY_XATTR_READ_ONLY),
         OPT2("--xattr-rw", "xattr-rw", OPTKEY_XATTR_READ_WRITE),
+        OPT2("--ctime-from-mtime", "ctime-from-mtime", OPTKEY_CTIME_FROM_MTIME),
         FUSE_OPT_END
     };
 
@@ -993,13 +1064,14 @@ int main(int argc, char *argv[])
     settings.create_policy = (getuid() == 0) ? CREATE_AS_USER : CREATE_AS_MOUNTER;
     settings.create_permchain = permchain_create();
     settings.chmod_policy = CHMOD_NORMAL;
-    settings.xattr_policy = XATTR_READ_ONLY;
+    settings.chmod_allow_x = 0;
+    settings.xattr_policy = XATTR_READ_WRITE;
     settings.mirrored_users_only = 0;
     settings.mirrored_users = NULL;
     settings.num_mirrored_users = 0;
     settings.mirrored_members = NULL;
     settings.num_mirrored_members = 0;
-
+    settings.ctime_from_mtime = 0;
     atexit(&atexit_func);
 
     /* Parse options */
